package com.javacorporate.xml.dbobj;

import com.javacorporate.db.DBException;
import com.javacorporate.dbobj.DBObject;
import java.io.*;
import java.util.Enumeration;
import java.net.URL;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.util.Vector;
import java.util.Enumeration;
import org.apache.xerces.parsers.*;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.HandlerBase;
import org.xml.sax.InputSource;

/**
 * Object to "wrap" a database object and allow import and
 * export of that object's data in XML format
 */

public class XMLDBObject {
	private DBObject myDBObject = null;
	private String thisClass = new String(getClass().getName() + ".");
	private Vector errorList = new Vector(3);

	/**
	 * How to we handle duplicate primary keys when importing? A = add new record
	 * U = update existing record
	 */
	private String dupHandling = new String("U");

	/**
	 * Default constructor
	 */
	public XMLDBObject()  {
		
		
	} /* XMLDBObject(String) */
	/**
	 * Default constructor
	 */
	public XMLDBObject(DBObject newDBObject)  {
		myDBObject = newDBObject;
		
	} /* XMLDBObject(DBObject) */
	/**
	 * Filename should NOT include an extension, as export writes
	 * a DTD and XML file 
	 */
	public void exportToFile(String fileName) throws DBException, IOException, FileNotFoundException {
		exportToFileXML(fileName);
		
	} /* exportToFile(String) */
	private void exportToFileXML(String fileName) throws FileNotFoundException, IOException, DBException {
		FileOutputStream fout = new FileOutputStream(fileName + ".xml", true);
		BufferedOutputStream bout = new BufferedOutputStream(fout);
		PrintStream XMLFileStream = new PrintStream(bout);

		exportToStreamXML(new PrintWriter(XMLFileStream));
		
		fout.close();
	}
/**
 * Export the records for the current selection in the specified database object into an XML format output stream
 * Creation date: (4/2/00 8:38:58 PM)
 * @param p java.io.PrintStream
 */
public void exportToStreamXML(PrintWriter p) throws DBException {
		String myName = new String(thisClass + "exportToStreamXML(PrintWriter)");
	
		int lineCount = 0;
		String fieldName = null;
		DBObject oneObj = null; 

		if (p == null) {
			throw new DBException(myName + ":PrintWriter cannot be null");
		}
		if (myDBObject == null) {
			throw new DBException(myName + ":Database object is not set - cannot export");
		}

		if (myDBObject.count() == 0) {
			return;
		}

		p.println("<DBObject>");
		p.println(indent(1) + "<ObjectName>" + myDBObject.getClass().getName()  + "</ObjectName>");
		p.println(indent(1) + "<TableName>" + myDBObject.getTableName() + "</TableName>");
		p.println(indent(1) + "<Data>");

		for (Enumeration rc = myDBObject.searchAndRetrieve().elements(); rc.hasMoreElements(); ) {	

			oneObj = (DBObject) rc.nextElement();

			p.println(indent(2) + "<DataRecord>");

			for (Enumeration e = myDBObject.getFieldList(); e.hasMoreElements(); ) {
				fieldName = (String) e.nextElement();

				p.println(indent(3) + "<Field>");
				p.println(indent(4) + "<FieldName>" + fieldName + "</FieldName>");
				p.println(indent(4) + "<FieldValue>" + oneObj.getField(fieldName) + "</FieldValue>");
				p.println(indent(3) + "</Field>");
			} /* for each field */
			lineCount++;

			p.println(indent(2) + "</DataRecord>");
			p.flush();
		
		} /* for each record */
		p.println(indent(1) + "</Data>");
		p.println("</DBObject>");

		p.flush();

}
	public DBObject getDBObject() throws DBException {
		return myDBObject;
	}
	public Vector getErrorList() {
		return errorList;
	}
	public void importFromFile(String fileName) throws DBException {
		String myName = new String(thisClass + "importFromFile(String)");
		
		try {
			XMLImportDocument xi = new XMLImportDocument(myDBObject);
			//SAXParserFactory spf = SAXParserFactory.newInstance();
			SAXParser sp = new SAXParser();
			
			sp.setContentHandler(xi);
			sp.setErrorHandler(xi);
			sp.parse(fileName + ".xml");
			//new InputSource(new FileInputStream(fileName + ".xml")), xi);
			errorList = xi.getErrorList();
			
		} catch (SAXParseException pe) {
			throw new DBException(myName + ":Parser Configuration Exception:" + pe.getMessage());
		} catch (SAXException se) {
			throw new DBException(myName + ":SAX Exception:" + se.getMessage());
		} catch (IOException ie) {
			throw new DBException(myName + ":IO Exception:" + ie.getMessage());
		}
	}
	public void importFromURL(String importURL) throws DBException {
		String myName = new String(thisClass + "importFromURL(String)");

		try {
			XMLImportDocument xi = new XMLImportDocument(myDBObject);
			SAXParser sp = new SAXParser();
			sp.setContentHandler(xi);
			sp.setErrorHandler(xi);
			System.out.println("URL:" + importURL);
			sp.parse(importURL);
			
		} catch (SAXParseException pe) {
			throw new DBException(myName + ":Parser Configuration Exception:" + pe.getMessage());
		} catch (SAXException se) {
			throw new DBException(myName + ":SAX Exception:" + se.getMessage());
		} catch (IOException ie) {
			throw new DBException(myName + ":IO Exception:" + ie.getMessage());
		}
		
	}
/**
 * Insert the method's description here.
 * Creation date: (4/4/00 11:27:40 AM)
 */
private String indent(int i) {
	StringBuffer tabString = new StringBuffer("");
	for (int j = 0; j < i; j++) {
		tabString.append("    ");
	}
	return tabString.toString();
}
/**
 * Insert the method's description here.
 * Creation date: (4/4/00 11:27:40 AM)
 */
void newMethod() {
}
	public void setDBObject(DBObject newDBObject) throws DBException {
		myDBObject = newDBObject;
	}
} /* XMLDBobject */
